"""
Fitting nonlinear genotype-phenotype maps
=========================================

Use a linear, logistic regression model to estimate the positive/negative effects
of mutations.
"""

# Imports
import matplotlib.pyplot as plt

from gpmap.simulate import MountFujiSimulation

from epistasis.models import EpistasisPowerTransform
from epistasis.pyplot import plot_power_transform

# The data
gpm = MountFujiSimulation.from_length(4, field_strength=-1, roughness=(-1,1))

# Initialize a model
model = EpistasisPowerTransform(lmbda=1, A=0, B=0)
model.add_gpm(gpm)

# Fit the model
model.fit()

fig, ax = plt.subplots(figsize=(2.5,2.5))
ax = plot_power_transform(model, ax=ax)
plt.show()



import numpy as np
import pandas as pd
from .stats import split_gpm, pearson


def k_fold(gpm, model, k=10):
    """Cross-validation using K-fold validation on a seer.
    """
    # Get index.
    idx = np.copy(gpm.index)

    # Shuffle index
    np.random.shuffle(idx)

    # Get subsets.
    subsets = np.array_split(idx, k)
    subsets_idx = np.arange(len(subsets))

    # Do k-fold
    scores = []
    for i in range(k):
        # Split index into train/test subsets
        train_idx = np.concatenate(np.delete(subsets, i))
        test_idx = subsets[i]

        # Split genotype-phenotype map
        train, test = split_gpm(gpm, idx=train_idx)

        # Fit model.
        model.add_gpm(train)
        model.fit()

        # Score validation set
        pobs = test.phenotypes
        pred = model.predict(X=test.genotypes)

        score = pearson(pobs, pred)**2
        scores.append(score)

    return scores


def holdout(gpm, model, size=1, repeat=1):
    """Validate a model by holding-out parts of the data.
    """
    train_scores = []
    test_scores = []

    model.add_gpm(gpm)
    X = model._X()

    for i in range(repeat):
        # Get index.
        idx = np.copy(gpm.index)

        # Shuffle index
        np.random.shuffle(idx)

        # Split model matriay to cross validate).
        train_idx = idx[:size]
        test_idx = idx[size:]
        train_X = X[train_idx, :]
        test_X = X[test_idx, :]

        # Train the model
        model.fit(X=train_X, y=gpm.phenotypes[train_idx])

        train_p = model.predict(X=train_X)
        train_s = pearson(train_p, gpm.phenotypes[train_idx])**2
        train_scores.append(train_s)

        # Test the model
        test_p = model.predict(X=test_X)
        test_s = pearson(test_p, gpm.phenotypes[test_idx])**2
        test_scores.append(test_s)


    return train_scores, test_scores



import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
import numpy as np

def corr(padd, pobs, perr=None, ax=None):
    """Make a correlation plot.
    """
    # Get 
    if ax is None:
        fig, ax = plt.subplots()
    else:
        fig = ax.get_figure()
        
    # Make correlation plot
    z = np.linspace(min(padd), max(padd), 2)
    
    # Add data to plot
    ax.plot(padd, pobs, 'o', color='k')
    ax.plot(z,z, '--', color='gray')
    
    # Set spines
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    
    # Set axis
    ax.axis('square')
    
    return fig, ax


def model(model):
    
    padd = model.Additive.predict("fit")
    pobs = model.gpm.phenotypes
    
    fig, ax = corr(padd, pobs)
    
    # Built a model line
    xmodel = np.linspace(min(padd), max(padd), 1000)
    ymodel = model.function(xmodel, *model.parameters.get_params())
    
    # Plot model
    ax.plot(xmodel, ymodel, '-', linewidth=2, color='r')
    
    return fig, ax
    
    



# from nose import tools
#
# from ..mapping import *
#
# def test_mutations_to_sites_1():
#     # Prepare data
#     mutations = {0:["A","V"], 1:["A", "V"]}
#     expected = [[0], [1], [2]]
#     # Test function
#     check = mutations_to_sites(1, mutations)
#     expected = [tuple(x) for x in expected]
#     check = [tuple(x) for x in check]
#     # Run tests
#     tools.assert_equals(set(expected), set(check))
#
# def test_mutations_to_sites_2():
#     # Prepare data
#     mutations = {0:["A","V"], 1:["A", "V"]}
#     expected = [[0], [1], [2], [1,2]]
#     # Test function
#     check = mutations_to_sites(2, mutations)
#     expected = [tuple(x) for x in expected]
#     check = [tuple(x) for x in check]
#     # Run tests
#     tools.assert_equals(set(expected), set(check))



import numpy as np

from .minimizer import Minimizer
from .ordinary import EpistasisNonlinearRegression
from epistasis.models import EpistasisLinearRegression
from epistasis.models.utils import (arghandler, FittingError)
from scipy.interpolate import UnivariateSpline
from lmfit import Parameter, Parameters


# -------------------- Minimizer object ------------------------

class SplineMinizer(Minimizer):
    """Spline Fitter.
    """
    def __init__(self, k=3, s=None):
        self.k = k
        self.s = s
        # Set initalize parameters to zero.
        self.parameters = Parameters()
        for i in range(self.k+1):
            self.parameters.add(name='c{}'.format(i), value=0)

    def _sorter(self, x, y=None, tol=1e-5):
        """sort x (and y) according to x values

        The spline call requires that x must be increasing. This function
        sorts x. If there are non-unique values, it adds a small numerical
        difference using tol.
        """
        # Copy x to leave it unchanged.
        x_ = np.copy(x)

        # Get non-unique terms
        idx = np.arange(len(x_))
        u, u_idx = np.unique(x, return_index=True)
        idx = np.delete(idx, u_idx)

        # Add noise to non-unique
        x_[idx] = x_[idx] + np.random.uniform(1,9.9, size=len(idx))*tol

        # Now sort x
        idx = np.argsort(x_)

        if y is None:
            return x_[idx]
        else:
            return x_[idx], y[idx]

    def function(self, x, *coefs):
        # Order of polynomial
        k = self.k

        # Number of coefficients
        n = k + 1

        # Knot positions
        t_arr = np.zeros(n*2, dtype=float)
        t_arr[:n] = min(x)
        t_arr[n:] = max(x)

        # Coefficients
        c_arr = np.zeros(n*2, dtype=float)
        c_arr[:n] = coefs

        # Build Spline function
        tck = [t_arr, c_arr, k]
        model = UnivariateSpline._from_tck(tck)

        # Return predicted function
        return model(x)

    def predict(self, x):
        return self._spline(x)

    def transform(self, x, y):
        ymodel = self.predict(x)
        return (y - ymodel) + x

    def fit(self, x, y):
        # Sort values for fit
        x_, y_ = self._sorter(x, y)

        # Fit spline.
        self._spline = UnivariateSpline(
            x=x_,
            y=y_,
            k=self.k,
            s=self.s
        )
        
        for i, coef in enumerate(self._spline.get_coeffs()):
            if 'c{}'.format(i) in self.parameters:
                self.parameters['c{}'.format(i)].value = coef
            else:
                raise FittingError('scipy.interpolate.UnivariateSpline '
                'fitting returned more parameters than\nexpected, likely'
                ' due to knots being added to closer fit the data.\nTry '
                'raising the value of `s` when initializing the spline '
                'model to prevent this.')
# -------------------- Minimizer object ------------------------

class EpistasisSpline(EpistasisNonlinearRegression):
    """Estimate nonlinearity in a genotype-phenotype map using a spline.

    Parameters
    ----------
    k : int
        order of spline.

    s : float, optional
        smoothness factor.
    """
    def __init__(self, k=3, s=None, model_type="global"):
        # Set atributes
        self.k = k
        self.s = s

        # Set up the function for fitting.
        self.minimizer = SplineMinizer(k=self.k, s=self.s)
        self.order = 1
        self.Xbuilt = {}

        # Construct parameters object
        self.set_params(model_type=model_type)

        # Store model specs.
        self.model_specs = dict(model_type=self.model_type)

        # Set up additive and high-order linear model
        self.Additive = EpistasisLinearRegression(
            order=1, model_type=self.model_type)



import numpy as np
from sklearn.linear_model import Lasso

from ..base import BaseModel, use_sklearn
from ..utils import arghandler

# Suppress an annoying error from scikit-learn
import warnings
warnings.filterwarnings(action="ignore", module="scipy",
                        message="^internal gelsd")


@use_sklearn(Lasso)
class EpistasisLasso(BaseModel):
    """A scikit-learn Lasso Regression class for discovering sparse
    epistatic coefficients.

    Methods are described in the following publication:
        Poelwijk FJ, Socolich M, and Ranganathan R. 'Learning the pattern of
        epistasis linking enotype and phenotype in a protein'.
        bioRxiv. (2017).

    Parameters
    ----------
    order : int
        order of epistasis

    model_type : str (default="global")
        model matrix type. See publication above for more information

    alpha : float
        Constant that multiplies the L1 term. Defaults to 1.0. alpha = 0 is
        equivalent to an ordinary least square, solved by the
        EpistasisLinearRegression object.

    precompute :
        Whether to use a precomputed Gram matrix to speed up calculations.
        If set to 'auto' let us decide. The Gram matrix can also be passed
        as argument. For sparse input this option is always True to preserve
        sparsity.

    max_iter : int
        The maximum number of iterations.

    tol : float
        The tolerance for the optimization: if the updates are smaller than
        tol, the optimization code checks the dual gap for optimality and
        continues until it is smaller than tol.

    warm_start : bool
        When set to True, reuse the solution of the previous call to fit as
        initialization, otherwise, just erase the previous solution.

    positive : bool
        When set to True, forces the coefficients to be positive.

    random_state : int
        The seed of the pseudo random number generator that selects a random
        feature to update. If int, random_state is the seed used by the random
        number generator; If RandomState instance, random_state is the random
        number generator; If None, the random number generator is the
        RandomState instance used by np.random. Used when
        selection == 'random'.

    selection : str
        If set to 'random', a random coefficient is updated every iteration
        rather than looping over features sequentially by default. This
        (setting to 'random') often leads to significantly faster convergence
        especially when tol is higher than 1e-4.
    """
    def __init__(
            self,
            order=1,
            model_type="global",
            alpha=1.0,
            precompute=False,
            max_iter=1000,
            tol=0.0001,
            warm_start=False,
            positive=False,
            random_state=None,
            selection='cyclic',
            **kwargs):
        # Set Linear Regression settings.
        self.fit_intercept = False
        self.normalize = False
        self.copy_X = True
        self.alpha = alpha
        self.precompute = precompute
        self.max_iter = max_iter
        self.tol = tol
        self.warm_start = warm_start
        self.positive = positive
        self.random_state = random_state
        self.selection = selection
        self.l1_ratio = 1.0

        self.set_params(model_type=model_type, order=order)
        self.Xbuilt = {}

        # Store model specs.
        self.model_specs = dict(
            order=self.order,
            model_type=self.model_type,
            **kwargs)

    def compression_ratio(self):
        """Compute the compression ratio for the Lasso regression
        """
        vals = self.epistasis.values
        zeros = vals[vals == 0]

        numer = len(zeros)
        denom = len(vals)
        return numer/denom

    @property
    def num_of_params(self):
        n = 0
        n += self.epistasis.n
        return n

    @arghandler
    def fit(self, X=None, y=None, **kwargs):
        # If a threshold exists in the data, pre-classify genotypes
        X = np.asfortranarray(X)
        self = super(self.__class__, self).fit(X, y)

        # Link coefs to epistasis values.
        self.epistasis.values = np.reshape(self.coef_, (-1,))
        return self

    def fit_transform(self, X=None, y=None, **kwargs):
        return self.fit(X=X, y=y, **kwargs)

    @arghandler
    def predict(self, X=None):
        X = np.asfortranarray(X)
        return super(self.__class__, self).predict(X)

    @arghandler
    def predict_transform(self, X=None, y=None):
        return self.predict(X=X)

    @arghandler
    def score(self, X=None, y=None):
        X = np.asfortranarray(X)
        return super(self.__class__, self).score(X, y)

    @property
    def thetas(self):
        return self.coef_

    @arghandler
    def hypothesis(self, X=None, thetas=None):
        return np.dot(X, thetas)

    @arghandler
    def hypothesis_transform(self, X=None, y=None, thetas=None):
        return self.hypothesis(X=X, thetas=thetas)

    @arghandler
    def lnlike_of_data(
            self,
            X=None, y=None,
            yerr=None,
            thetas=None):

        # Calculate y from model.
        ymodel = self.hypothesis(X=X, thetas=thetas)

        # Return the likelihood of this model (with an L1 prior)
        return (- 0.5 * np.log(2 * np.pi * yerr**2) -
                (0.5 * ((y - ymodel)**2 / yerr**2)) -
                (self.alpha * sum(abs(thetas))))

    @arghandler
    def lnlike_transform(
            self,
            X=None,
            y=None,
            yerr=None,
            lnprior=None,
            thetas=None):
        # Update likelihood.
        lnlike = self.lnlike_of_data(X=X, y=y, yerr=yerr, thetas=thetas)
        return lnlike + lnprior



import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
import numpy as np

def corr(y_pred, y_obs, fig=None, ax=None, figsize=None):
    """Make a correlation plot.
    """
    if fig is None:
        fig = plt.figure(figsize=figsize)
    if ax is None:
        ax = fig.add_subplot(111)
    # Make correlation plot
    z = np.linspace(min(y_pred), max(y_pred),2)
    ax.plot(y_pred, y_obs, '.')
    ax.plot(z,z, '--', color="gray")
    ax.spines["top"].set_visible(False)
    ax.spines["right"].set_visible(False)
    ax.axis("equal")
    return fig, ax

def resid(y_pred, y_obs, fig=None, ax=None, figsize=None):
    """Make a residual plot between y_obs and y_pred (i.e. y_obs - y_pred).
    """
    if fig is None:
        fig = plt.figure(figsize=figsize)
    if ax is None:
        ax = fig.add_subplot(111)
    residuals = y_obs - y_pred
    # Make residual plot
    ax.plot(residuals, '.')
    ax.hlines(0,min(y_pred), max(y_pred), linestyle="--", color="gray")
    ax.spines["top"].set_visible(False)
    ax.spines["right"].set_visible(False)
    return fig, ax

def rhist(y_pred, y_obs, fig=None, ax=None, figsize=None):
    """Make a horizontal histogram for residuals between y_obs and y_pred.
    """
    if fig is None:
        fig = plt.figure(figsize=figsize)
    if ax is None:
        ax = fig.add_subplot(111)
    residuals = y_obs - y_pred
    # Make histogram plot
    ax.hist(residuals, bins=10, orientation="horizontal")
    ax.hlines(0,min(y_pred), max(y_pred), linestyle="--", color="gray")
    ax.spines["top"].set_visible(False)
    ax.spines["right"].set_visible(False)
    return fig, ax

def corr_resid(y_pred, y_obs, figsize=None):
    """Make a correlation and residuals plot using matplotlib.

    Parameters
    ----------
    y_pred : array
        predicted values (will be placed on x axis)
    y_obs : array
        observed values (will be placed on y axis).

    Returns
    -------
    fig : Figure object
        Figure object
    ax_c : Axes object
        subplot with correlation plot drawn
    ax_r : Axes object
        subplot with residual plot drawn
    """
    # Initialize the grid.
    fig = plt.figure(figsize=figsize)
    gs = gridspec.GridSpec(2, 1, height_ratios=[9,1])
    ax_c = fig.add_subplot(gs[0, :])
    ax_r = fig.add_subplot(gs[1 ,:], sharex=ax_c)

    # build subplots
    corr(y_pred, y_obs, fig=fig, ax=ax_c)
    resid(y_pred, y_obs, fig=fig, ax=ax_r)
    return fig, ax_c, ax_r

def corr_resid_rhist(y_pred, y_obs, figsize=(3,4)):
    """Make a correlation, residuals, histogram plot using matplotlib.

    Parameters
    ----------
    y_pred : array
        predicted values (will be placed on x axis)
    y_obs : array
        observed values (will be placed on y axis).

    Returns
    -------
    fig : Figure object
        Figure object
    ax_c : Axes object
        subplot
    ax_r : Axes object
        subplot with residual plot drawn
    ax_h : Axes object
        subplot with histogram plot drawn
    """
    fig = plt.figure(figsize=figsize)
    gs = gridspec.GridSpec(2,2, height_ratios=[4,1], width_ratios=[4,1])
    # Construct grid.
    ax_c = fig.add_subplot(gs[0:1, 0:1])
    ax_r = fig.add_subplot(gs[1: , 0:1], sharex=ax_c)
    ax_h = fig.add_subplot(gs[1: , 1: ], sharey=ax_r)

    corr(y_pred, y_obs, fig=fig, ax=ax_c)
    resid(y_pred, y_obs, fig=fig, ax=ax_r)
    rhist(y_pred, y_obs, fig=fig, ax=ax_h)
    return fig, ax_c, ax_r, ax_h



# By Jake VanderPlas
# License: BSD-style

import matplotlib.pyplot as plt
import numpy as np
from epistasis.utils import Bunch

def discrete_cmap(N, base_cmap=None):
    """Create an N-bin discrete colormap from the specified input map.

    Borrowed from Jake VanderPlas's gist found here:
    https://gist.github.com/jakevdp/91077b0cae40f8f8244a
    """
    # Note that if base_cmap is a string or None, you can simply do
    #    return plt.cm.get_cmap(base_cmap, N)
    # The following works for string, None, or a colormap instance:
    base = plt.cm.get_cmap(base_cmap)
    color_list = base(np.linspace(0, 1, N))
    cmap_name = base.name + str(N)
    return base.from_list(cmap_name, color_list, N)



import numpy as np
import matplotlib.pyplot as plt
from matplotlib.path import Path
from matplotlib import patches
import warnings

def fraction_explained(fraction_explained, color_vector=None, num_bins=1000,lw=0.25):
    """
    Plot a square "pie" chart where each box is colored according to how often it
    occurs in the data set.

    Parameters
    ----------
    """
    # Normalize fx_vector so it adds up to 1.0
    internal_fx_vector = np.array(np.copy(fraction_explained))
    if np.sum(internal_fx_vector) != 1.0:
        warnings.warn("fx_vector does not add up to 1")
        internal_fx_vector = internal_fx_vector/np.sum(internal_fx_vector)

    # Create a color vector or grab the one off the command line
    if color_vector is None:
        # Prepare an cycle of colors
        order = len(fraction_explained)
        prop_cycle = plt.rcParams['axes.prop_cycle']
        color_vector = prop_cycle.by_key()['color']
        color_scalar = int(order / len(color_vector))  + 1
        color_vector *= color_scalar
    else:
        if len(fx_vector) > len(color_vector):
            err = "len(color_vector) must be >= len(fx_vector)\n"
            raise ValueError(err)

    # Discretize the input vector with appropriately scaled
    side_length = np.int(np.round(np.sqrt(num_bins),0))
    num_bins = side_length*side_length
    fx_vector_int = np.cumsum(np.round(internal_fx_vector*num_bins,0)).astype(int)

    # Generate the plot
    fig = plt.figure()
    ax = fig.add_subplot(111)

    # path codes for drawing the boxes
    box_codes = [Path.MOVETO,Path.LINETO,Path.LINETO,Path.LINETO,Path.CLOSEPOLY]

    # Go through each bin and color appropriately
    current_index = 0
    total_counter = 0
    for i in range(side_length-1,-1,-1):
        for j in range(side_length):
            if total_counter >= fx_vector_int[current_index]:
                current_index += 1
            # kind of a hack.  last entry sometimes has round error and doesn't get given a color.
            # use last entry to fill in.
            if current_index >= len(fraction_explained):
                current_index -= 1
            # Draw box
            verts = [(j,i),(j+1,i),(j+1,i+1),(j,i+1),(j, i),]
            path = Path(verts, box_codes)
            patch = patches.PathPatch(path, facecolor=color_vector[current_index], lw=lw)
            ax.add_patch(patch)
            total_counter += 1

    # Clean up plot
    ax.axis('equal')
    ax.axis('off')
    ax.set_xlim(-1,side_length+1)
    ax.set_ylim(-1,side_length+1)
    return fig, ax



import numpy as np
import pandas as pd

from epistasis.models.base import BaseModel, use_sklearn
from epistasis.models.utils import arghandler


# Use if inheriting from a scikit-learn class
#@use_sklearn()
class EpistasisModel(BaseModel):
    """Epistasis model template.
    """
    def __init__(self, ):
        pass

    @property
    def num_of_params(self):
        pass

    #@arghandler
    def fit(self, X=None, y=None, **kwargs):
        pass

    def fit_transform(self, X=None, y=None, **kwargs):
        pass

    def predict(self, X=None):
        pass

    def predict_transform(self, X=None, y=None, **kwargs):
        pass

    def hypothesis(self, X=None, thetas=None):
        pass

    def hypothesis_transform(self, X=None, y=None, thetas=None):
        pass

    def lnlike_of_data(
        self,
        X=None,
        y=None,
        yerr=None,
        thetas=None):

        pass

    def lnlike_transform(
        self,
        X=None,
        y=None,
        yerr=None,
        lnprior=None,
        thetas=None):
        pass



# External imports
import unittest
import pytest

import numpy as np
from gpmap import GenotypePhenotypeMap

# Module to test
from ..ridge import EpistasisRidge


@pytest.fixture
def gpm():
    """Create a genotype-phenotype map"""
    wildtype = "000"
    genotypes = ["000", "001", "010", "100", "011", "101", "110", "111"]
    phenotypes = [0.1,   0.1,   0.5,   0.4,   0.2,   0.8,   0.5,   1.0]
    stdeviations = 0.1
    return GenotypePhenotypeMap(wildtype, genotypes, phenotypes,
                                stdeviations=stdeviations)


class TestEpistasisRidge(object):

    order = 3

    def test_init(self, gpm):
        model = EpistasisRidge(order=self.order, model_type="local")
        model.add_gpm(gpm)

        # Checks
        check1 = model.order
        check2 = model.model_type
        assert check1 == self.order
        assert check2 == "local"

    def test_fit(self, gpm):
        model = EpistasisRidge(order=self.order, model_type="local")
        model.add_gpm(gpm)
        model.fit()
        # Checks
        check1 = hasattr(model, "Xbuilt")
        check2 = hasattr(model, "coef_")
        check3 = hasattr(model, "epistasis")

        # Tests
        assert check1 is True
        assert check2 is True
        assert check3 is True
        assert "fit" in model.Xbuilt


    def test_predict(self, gpm):
        model = EpistasisRidge(order=self.order, model_type="local")
        model.add_gpm(gpm)
        model.fit()
        check1 = model.predict(X='fit')

        # Tests
        assert "predict" in model.Xbuilt

    def test_score(self, gpm):
        model = EpistasisRidge(order=self.order, model_type="local", alpha=0.1)
        model.add_gpm(gpm)
        model.fit()
        score = model.score()
        # Tests
        assert score >= 0
        assert score <= 1

    def test_hypothesis(self, gpm):
        model = EpistasisRidge(order=self.order, model_type="local")
        model.add_gpm(gpm)
        model.fit()
        # Checks
        check1 = model.hypothesis(thetas=model.coef_)
        # Tests
        assert True

    def test_lnlikelihood(self, gpm):
        model = EpistasisRidge(order=self.order, model_type="local")
        model.add_gpm(gpm)
        model.fit()

        # Calculate lnlikelihood
        lnlike = model.lnlikelihood()
        assert lnlike.dtype == float



import numpy as np

DTYPE = np.int

cimport numpy as np

ctypedef np.int_t DTYPE_t

def build_model_matrix(np.ndarray encoding_vectors, np.ndarray sites):

    cdef int n = encoding_vectors.shape[0]
    cdef int m = sites.shape[0]
    cdef int i, j, k, l, element
    cdef np.ndarray matrix = np.ones([n, m], dtype=DTYPE)

    # Interate over rows
    for i in range(n):

        # Iterate over cols
        for j in range(m):
            element = 1
            # Product of sites in encoding vector
            l = sites[j].shape[0]
            for k in range(l):
                element = element * encoding_vectors[i, sites[j][k]]

            # Multiply these elements
            matrix[i,j] = element

    return matrix



"""
Handling dead phenotypes
========================

Use a linear, logistic regression model to estimate the positive/negative effects
of mutations.
"""

# Imports
import matplotlib.pyplot as plt

from gpmap import GenotypePhenotypeMap
from epistasis.models import EpistasisLogisticRegression
from epistasis.pyplot import plot_coefs

# The data
wildtype = "000"
genotypes = ['000', '001', '010', '011', '100', '101', '110', '111']
phenotypes = [ 0.366, -0.593,  1.595, -0.753,  0.38 ,  1.296,  1.025, -0.519]
gpm = GenotypePhenotypeMap(wildtype, genotypes, phenotypes)

# Threshold
threshold = 1.0

# Initialize a model
model = EpistasisLogisticRegression(threshold=threshold)
model.add_gpm(gpm)

# Fit the model
model.fit()

fig, ax = plot_coefs(model, figsize=(1,3))
plt.show()



s__doc__ = """Submodule with various classes for generating/simulating
genotype-phenotype maps."""

# ------------------------------------------------------------
# Imports
# ------------------------------------------------------------

import numpy as np
from gpmap.gpm import GenotypePhenotypeMap

# local imports
from epistasis.matrix import get_model_matrix
from epistasis.simulate.base import BaseSimulation

# ------------------------------------------------------------
# ArtificialMap object can be used to quickly generating a toy
# space for testing the EpistasisModels
# ------------------------------------------------------------


class LinearSimulation(BaseSimulation):
    """Construct an genotype-phenotype from linear building blocks and
    epistatic coefficients.

    Example
    -------
    Phenotype = b0 + b1 + b2 + b3 + b12 + b13 + b13 + b123

    Parameters
    ---------
    wildtype : str
        Wildtype genotype
    mutations : dict
        Mapping for each site to its alphabet
    order : int
        Order of epistasis in simulated genotype-phenotype map
    betas : array-like
        values of epistatic coefficients (must be positive for this function
        to work. Log is taken)
    model_type : str
        Use a local or global (i.e. Walsh space) epistasis model to construct
        phenotypes
    """

    def __init__(self, wildtype, mutations,
                 model_type='global',
                 **kwargs):
        # Construct epistasis mapping objects (empty)
        super(LinearSimulation, self).__init__(
            wildtype,
            mutations,
            **kwargs)
        self.model_type = model_type

    def build(self):
        """ Build the phenotype map from epistatic interactions. """
        X = self.add_X()
        # Get model type:
        self.data['phenotypes'] = np.dot(X, self.epistasis.values)



import numpy as np
import pandas as pd

# Scikit-learn classifiers
from sklearn.preprocessing import binarize

from epistasis.mapping import EpistasisMap
from epistasis.models.base import BaseModel, use_sklearn
from epistasis.models.utils import (XMatrixException, arghandler)

from epistasis.models.linear import EpistasisLinearRegression

from gpmap import GenotypePhenotypeMap


class EpistasisClassifierMixin:
    """A Mixin class for epistasis classifiers
    """
    def _fit_additive(self, X=None, y=None):
        # Construct an additive model.
        self.Additive = EpistasisLinearRegression(
            order=1, model_type=self.model_type)

        self.Additive.add_gpm(self.gpm)

        # Prepare a high-order model
        self.Additive.epistasis = EpistasisMap(
            sites=self.Additive.Xcolumns,
        )

        # Fit the additive model and infer additive phenotypes
        self.Additive.fit(X=X, y=y)
        return self

    def _fit_classifier(self, X=None, y=None):
        # This method builds x and y from data.
        add_coefs = self.Additive.epistasis.values
        add_X = self.Additive._X(data=X)

        # Project X into padd space.
        X = add_X * add_coefs

        # Label X.
        y = binarize(y.reshape(1, -1), self.threshold)[0]
        self.classes = y

        # Fit classifier.
        super().fit(X=X, y=y)
        return self

    def fit_transform(self, X=None, y=None, **kwargs):
        self.fit(X=X, y=y, **kwargs)
        ypred = self.predict(X=X)

        # Transform map.
        gpm = GenotypePhenotypeMap.read_dataframe(
            dataframe=self.gpm.data[ypred==1],
            wildtype=self.gpm.wildtype,
            mutations=self.gpm.mutations
        )
        return gpm

    def predict(self, X=None):
        Xadd = self.Additive._X(data=X)
        X = Xadd * self.Additive.epistasis.values
        return super().predict(X=X)

    def predict_transform(self, X=None, y=None):
        x = self.predict(X=X)
        y[x <= 0.5] = self.threshold
        return y

    def predict_log_proba(self, X=None):
        Xadd = self.Additive._X(data=X)
        X = Xadd * self.Additive.epistasis.values
        return super().predict_log_proba(X)

    def predict_proba(self, X=None):
        Xadd = self.Additive._X(data=X)
        X = Xadd * self.Additive.epistasis.values
        return super().predict_proba(X=X)



import numpy as np
import pandas as pd

# Scikit-learn classifiers
from sklearn.mixture import GaussianMixture
from sklearn.preprocessing import binarize

from epistasis.mapping import EpistasisMap
from epistasis.models.base import BaseModel, use_sklearn
from epistasis.models.utils import (XMatrixException, arghandler)

from epistasis.models.linear import EpistasisLinearRegression

from gpmap import GenotypePhenotypeMap

from .base import EpistasisClassifierMixin

@use_sklearn(GaussianMixture)
class EpistasisGaussianMixture(EpistasisClassifierMixin, BaseModel):
    """Logistic regression for estimating epistatic interactions that lead to
    nonviable phenotypes. Useful for predicting viable/nonviable phenotypes.

    Parameters
    ----------
    order : int
        order of epistasis model

    model_type : str (default="global")
        type of model matrix to use. "global" defines epistasis with respect to
        a background-averaged "genotype-phenotype". "local" defines epistasis
        with respect to the wildtype genotype.
    """
    def __init__(
        self,
        n_components=1,
        model_type="global",
        **kwargs):

        super(self.__class__, self).__init__(n_components=n_components, **kwargs)
        self.model_type = model_type
        self.order = 1
        self.Xbuilt = {}

        # Store model specs.
        self.model_specs = dict(
            model_type=self.model_type,
            **kwargs)

    @arghandler
    def lnlike_of_data(self, X=None, y=None, yerr=None, thetas=None):
        pass

    @arghandler
    def lnlike_transform(
        self,
        X=None,
        y=None,
        yerr=None,
        lnprior=None,
        thetas=None):
        pass

    @arghandler
    def hypothesis(self, X=None, thetas=None):
        pass

    def hypothesis_transform(self, X=None, y=None, thetas=None):
        pass

    @property
    def thetas(self):
        pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Note: To use the 'upload' functionality of this file, you must:
#   $ pip install twine

import io
import os
import sys
from shutil import rmtree

from setuptools import find_packages, setup, Command, Extension

# Package meta-data.
NAME = 'epistasis'
DESCRIPTION = 'A Python API for estimating statistical high-order epistasis in genotype-phenotype maps.'
URL = 'https://github.com/harmslab/epistasis'
EMAIL = 'zachsailer@gmail.com'
AUTHOR = 'Zachary R. Sailer'
REQUIRES_PYTHON = '>=3.3.0'
VERSION = None

# What packages are required for this module to be executed?
REQUIRED = [
    "cython",
    "numpy>=1.15.2",
    "pandas>=0.24.2",
    "scikit-learn>=0.20.0",
    "scipy>=1.1.0",
    "emcee>=2.2.1",
    "lmfit>=0.9.11",
    "matplotlib>=3.0.0",
    "gpmap>=0.6.0",
]

# Hanlding a Cython extension is a pain! Need to
# import numpy before it's installed... Used this
# Stackoverflow solution:
# https://stackoverflow.com/a/42163080

try:
    from Cython.setuptools import build_ext
except:
    # If we couldn't import Cython, use the normal setuptools
    # and look for a pre-compiled .c file instead of a .pyx file
    from setuptools.command.build_ext import build_ext
    extension = Extension("epistasis.matrix_cython", ["epistasis/matrix_cython.c"])
else:
    # If we successfully imported Cython, look for a .pyx file
    extension = Extension("epistasis.matrix_cython", ["epistasis/matrix_cython.pyx"])

class CustomBuildExtCommand(build_ext):
    """build_ext command for use when numpy headers are needed."""
    def run(self):

        # Import numpy here, only when headers are needed
        import numpy

        # Add numpy headers to include_dirs
        self.include_dirs.append(numpy.get_include())

        # Call original build_ext command
        build_ext.run(self)


# The rest you shouldn't have to touch too much :)
# ------------------------------------------------
# Except, perhaps the License and Trove Classifiers!
# If you do change the License, remember to change the Trove Classifier for that!

here = os.path.abspath(os.path.dirname(__file__))

# Import the README and use it as the long-description.
# Note: this will only work if 'README.md' is present in your MANIFEST.in file!
with io.open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = '\n' + f.read()

# Load the package's __version__.py module as a dictionary.
about = {}
if not VERSION:
    with open(os.path.join(here, NAME, '__version__.py')) as f:
        exec(f.read(), about)
else:
    about['__version__'] = VERSION


# Where the magic happens:
setup(
    name=NAME,
    version=about['__version__'],
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type='text/markdown',
    author=AUTHOR,
    author_email=EMAIL,
    python_requires=REQUIRES_PYTHON,
    url=URL,
    packages=find_packages(exclude=('tests',)),
    ext_modules=[extension],
    install_requires=REQUIRED,
    extras_require = {
        'test': ['pytest'],
    },
    include_package_data=True,
    license='UNLICENSE',
    classifiers=[
      'Development Status :: 3 - Alpha',
      'Intended Audience :: Science/Research',
      'Programming Language :: Python :: 3.4',
      'Programming Language :: Python :: 3.5',
      'Programming Language :: Python :: 3.6',
    ],
    keywords='epistasis high-order genetics genotype-phenotype-maps',
    cmdclass = {'build_ext': CustomBuildExtCommand},
)



"""

A module for simulating genotype-phenotype maps with high-order epistasis.

"""
# Hide the base class
__all__ = [
    "linear",
    "nonlinear"
]

# Load all Simulation classes on init.
from .linear import LinearSimulation
from .power import PowerScaleSimulation



"""
High-order epistasis model
==========================

Estimate high-order epistatic coefficients in arbitrary genotype-phenotype maps.
A linear epistasis model fits high-order interaction terms to capture
variation in phenotype.
"""
# Imports
import matplotlib.pyplot as plt

from gpmap import GenotypePhenotypeMap
from epistasis.models import EpistasisLinearRegression
from epistasis.pyplot import plot_coefs


# The data
wildtype = "000"
genotypes = ['000', '001', '010', '011', '100', '101', '110', '111']
phenotypes = [ 0.366, -0.593,  1.595, -0.753,  0.38 ,  1.296,  1.025, -0.519]
gpm = GenotypePhenotypeMap(wildtype, genotypes, phenotypes)

# Initialize a model
model = EpistasisLinearRegression(order=3)
model.add_gpm(gpm)

# Fit the model
model.fit()

fig, ax = plot_coefs(model, figsize=(2,3))
plt.show()



import matplotlib.cm as cm
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
import numpy as np

def corr(padd, pobs, perr=None, ax=None, color='k'):
    """Make a correlation plot.
    """
    # Get
    if ax is None:
        fig, ax = plt.subplots()
    else:
        fig = ax.get_figure()

    # Make correlation plot
    z = np.linspace(min(padd), max(padd), 2)

    # Add data to plot
    ax.scatter(padd, pobs, marker='o', c=color, cmap="coolwarm", vmin=0, vmax=1)
    ax.plot(z,z, '--', color='gray')
    
    # Set spines
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)

    # Set axis
    ax.axis('square')
    return fig, ax

def model(model):

    # Get Padd vs Pobs datapoints and classes
    padd = model.Model.Additive.predict(X="obs")
    pprob = model.Classifier.predict_proba(X="obs")
    pobs = model.gpm.phenotypes

    # Build a color array for the two classes
    #color = np.empty(len(pclass), dtype=str)
    #color[pclass==0] = "r"
    #color[pclass==1] = "k"

    # Construct the correlation plot between Pobs and Padd
    fig, ax = corr(padd, pobs, color=pprob[:,1])

    # Built a model line
    xmodel = np.linspace(min(padd), max(padd), 1000)
    ymodel = model.Model.function(xmodel, *model.Model.parameters.get_params())

    # Plot model
    ax.plot(xmodel, ymodel, '-', linewidth=4, color='k', alpha=0.5)
    ylim = ax.get_ylim()
    xlim = ax.get_xlim()

    # Plot the "dead" region on the Pobs axis.
    ybox1 = [model.threshold, model.threshold]
    ybox2 = [ylim[0], ylim[0]]
    xbox = [xlim[0], xlim[1]]
    ax.fill_between(xbox, ybox1, ybox2, alpha=0.1, color='C0')
    return fig, ax



"""\
A Python API for modeling statistical, high-order epistasis in genotype-phenotype maps.
This library provides methods for:

    1. Decomposing genotype-phenotype maps into high-order epistatic interactions
    2. Finding nonlinear scales in the genotype-phenotype map
    3. Calculating the contributions of different epistatic orders
    4. Estimating the uncertainty of epistatic coefficients amd
    5. Interpreting the evolutionary importance of high-order interactions.

For more information about the epistasis models in this library, see our Genetics paper:

    `Sailer, Z. R., & Harms, M. J. (2017). "Detecting High-Order Epistasis in Nonlinear Genotype-Phenotype Maps." Genetics, 205(3), 1079-1088.`_

.. _`Sailer, Z. R., & Harms, M. J. (2017). "Detecting High-Order Epistasis in Nonlinear Genotype-Phenotype Maps." Genetics, 205(3), 1079-1088.`: http://www.genetics.org/content/205/3/1079

Currently, this package works only as an API and there is no command-line
interface. Instead, we encourage you use this package inside `Jupyter notebooks`_ .
"""
from .__version__ import __version__

# from . import models
# from . import simulate
# from . import sampling
# from . import pyplot



import inspect
import numpy as np
import pandas as pd
from functools import wraps
from epistasis.matrix import get_model_matrix
from epistasis.mapping import EpistasisMap

from gpmap.utils import genotypes_to_binary

class XMatrixException(Exception):
    """Exception Subclass for X matrix errors."""


class FittingError(Exception):
    """Exception Subclass for X matrix errors."""

def arghandler(method):
    """Points methods to argument handlers. Assumes each argument has a
    corresponding method attached to the object named "_{argument}". These
    methods given default values to arguments.

    Ignores self and kwargs
    """
    @wraps(method)
    def inner(self, *args, **kwargs):
        # Get method name
        name = method.__name__

        # Inspect function for arguments to update.
        out = inspect.signature(method)

        # Construct kwargs from signature.
        kws = {key: val.default for key, val in out.parameters.items()}
        kws.pop('self')

        # Try to remove kwargs.
        try:
            kws.pop('kwargs')
        except:
            pass

        # Update kwargs with user specified kwargs.
        kws.update(**kwargs)

        # Handle each argument
        for arg in kws:
            # Get handler function.
            handler_name = "_{}".format(arg)
            handler = getattr(self, handler_name)
            kws[arg] = handler(data=kws[arg], method=name)

        return method(self, **kws)
    return inner
